import numpy as np
from tqdm import tqdm
import os
from typing import Union, Optional, List, Dict, Set, Any, Tuple, Literal
import logging
from copy import deepcopy
import pandas as pd

from .utils.misc_utils import compute_mdhash_id, NerRawOutput, TripleRawOutput

logger = logging.getLogger(__name__)

class EmbeddingStore:
    def __init__(self, embedding_model, db_filename, batch_size, namespace):
        """
        Initializes the class with necessary configurations and sets up the working directory.

        Parameters:
        embedding_model: The model used for embeddings.
        db_filename: The directory path where data will be stored or retrieved.
        batch_size: The batch size used for processing.
        namespace: A unique identifier for data segregation.

        Functionality:
        - Assigns the provided parameters to instance variables.
        - Checks if the directory specified by `db_filename` exists.
          - If not, creates the directory and logs the operation.
        - Constructs the filename for storing data in a parquet file format.
        - Calls the method `_load_data()` to initialize the data loading process.
        """
        self.embedding_model = embedding_model
        self.batch_size = batch_size
        self.namespace = namespace

        if not os.path.exists(db_filename):
            logger.info(f"Creating working directory: {db_filename}")
            os.makedirs(db_filename, exist_ok=True)

        self.filename = os.path.join(
            db_filename, f"vdb_{self.namespace}.parquet"
        )
        self._load_data()

    def get_missing_string_hash_ids(self, texts: List[str]):
        nodes_dict = {}

        for text in texts:
            nodes_dict[compute_mdhash_id(text, prefix=self.namespace + "-")] = {'content': text}

        # Get all hash_ids from the input dictionary.
        all_hash_ids = list(nodes_dict.keys())
        if not all_hash_ids:
            return  {}

        existing = self.hash_id_to_row.keys()

        # Filter out the missing hash_ids.
        missing_ids = [hash_id for hash_id in all_hash_ids if hash_id not in existing]
        texts_to_encode = [nodes_dict[hash_id]["content"] for hash_id in missing_ids]

        return {h: {"hash_id": h, "content": t} for h, t in zip(missing_ids, texts_to_encode)}

    def insert_strings(self, texts: List[str]):
        nodes_dict = {}

        for text in texts:
            nodes_dict[compute_mdhash_id(text, prefix=self.namespace + "-")] = {'content': text}

        # Get all hash_ids from the input dictionary.
        all_hash_ids = list(nodes_dict.keys())
        if not all_hash_ids:
            return  # Nothing to insert.

        existing = self.hash_id_to_row.keys()

        # Filter out the missing hash_ids.
        missing_ids = [hash_id for hash_id in all_hash_ids if hash_id not in existing]

        logger.info(
            f"Inserting {len(missing_ids)} new records, {len(all_hash_ids) - len(missing_ids)} records already exist.")

        if not missing_ids:
            return  {}# All records already exist.

        # Prepare the texts to encode from the "content" field.
        texts_to_encode = [nodes_dict[hash_id]["content"] for hash_id in missing_ids]

        missing_embeddings = self.embedding_model.batch_encode(texts_to_encode)

        self._upsert(missing_ids, texts_to_encode, missing_embeddings)

    def _load_data(self):
        if os.path.exists(self.filename):
            df = pd.read_parquet(self.filename)
            self.hash_ids, self.texts, self.embeddings = df["hash_id"].values.tolist(), df["content"].values.tolist(), df["embedding"].values.tolist()
            self.hash_id_to_idx = {h: idx for idx, h in enumerate(self.hash_ids)}
            self.hash_id_to_row = {
                h: {"hash_id": h, "content": t}
                for h, t in zip(self.hash_ids, self.texts)
            }
            self.hash_id_to_text = {h: self.texts[idx] for idx, h in enumerate(self.hash_ids)}
            self.text_to_hash_id = {self.texts[idx]: h  for idx, h in enumerate(self.hash_ids)}
            assert len(self.hash_ids) == len(self.texts) == len(self.embeddings)
            logger.info(f"Loaded {len(self.hash_ids)} records from {self.filename}")
        else:
            self.hash_ids, self.texts, self.embeddings = [], [], []
            self.hash_id_to_idx, self.hash_id_to_row = {}, {}

    def _save_data(self):
        data_to_save = pd.DataFrame({
            "hash_id": self.hash_ids,
            "content": self.texts,
            "embedding": self.embeddings
        })
        data_to_save.to_parquet(self.filename, index=False)
        self.hash_id_to_row = {h: {"hash_id": h, "content": t} for h, t, e in zip(self.hash_ids, self.texts, self.embeddings)}
        self.hash_id_to_idx = {h: idx for idx, h in enumerate(self.hash_ids)}
        self.hash_id_to_text = {h: self.texts[idx] for idx, h in enumerate(self.hash_ids)}
        self.text_to_hash_id = {self.texts[idx]: h for idx, h in enumerate(self.hash_ids)}
        logger.info(f"Saved {len(self.hash_ids)} records to {self.filename}")

    def _upsert(self, hash_ids, texts, embeddings):
        self.embeddings.extend(embeddings)
        self.hash_ids.extend(hash_ids)
        self.texts.extend(texts)

        logger.info(f"Saving new records.")
        self._save_data()

    def delete(self, hash_ids):
        indices = []

        for hash in hash_ids:
            indices.append(self.hash_id_to_idx[hash])

        sorted_indices = np.sort(indices)[::-1]

        for idx in sorted_indices:
            self.hash_ids.pop(idx)
            self.texts.pop(idx)
            self.embeddings.pop(idx)

        logger.info(f"Saving record after deletion.")
        self._save_data()

    def get_row(self, hash_id):
        return self.hash_id_to_row[hash_id]

    def get_hash_id(self, text):
        return self.text_to_hash_id[text]

    def get_rows(self, hash_ids, dtype=np.float32):
        if not hash_ids:
            return {}

        results = {id : self.hash_id_to_row[id] for id in hash_ids}

        return results

    def get_all_ids(self):
        return deepcopy(self.hash_ids)

    def get_all_id_to_rows(self):
        return deepcopy(self.hash_id_to_row)

    def get_all_texts(self):
        return set(row['content'] for row in self.hash_id_to_row.values())

    def get_embedding(self, hash_id, dtype=np.float32) -> np.ndarray:
        return self.embeddings[self.hash_id_to_idx[hash_id]].astype(dtype)
    
    def get_embeddings(self, hash_ids, dtype=np.float32) -> list[np.ndarray]:
        if not hash_ids:
            return []

        indices = np.array([self.hash_id_to_idx[h] for h in hash_ids], dtype=np.intp)
        embeddings = np.array(self.embeddings, dtype=dtype)[indices]

        return embeddings


from typing import Dict, List, Union, Any, Optional


from ..prompts.prompt_template_manager import PromptTemplateManager
from .logging_utils import get_logger
from ..llm.request_llm import SiliconFlowLLM

logger = get_logger(__name__)



def merge_elements_with_same_first_line(elements, prefix='Wikipedia Title: '):
    merged_dict = {}

    # Iterate through each element in the list
    for element in elements:
        # Split the element into lines and get the first line
        lines = element.split('\n')
        first_line = lines[0]

        # Check if the first line is already a key in the dictionary
        if first_line in merged_dict:
            # Append the current element to the existing value
            merged_dict[first_line] += "\n" + element.split(first_line, 1)[1].strip('\n')
        else:
            # Add the current element as a new entry in the dictionary
            merged_dict[first_line] = prefix + element

    # Extract the merged elements from the dictionary
    merged_elements = list(merged_dict.values())
    return merged_elements


def reason_step(dataset, prompt_template_manager: PromptTemplateManager, query: str, passages: list, thoughts: list, llm_client: SiliconFlowLLM):
    """
    Given few-shot samples, query, previous retrieved passages, and previous thoughts, generate the next thought with OpenAI models. The generated thought is used for further retrieval step.
    :return: next thought
    """

    prompt_user = ''
    if dataset in ['hotpotqa', 'hotpotqa_train']:
        passages = merge_elements_with_same_first_line(passages)
    for passage in passages:
        prompt_user += f'{passage}\n\n'
    prompt_user += f'Question: {query}\nThought:' + ' '.join(thoughts)

    
    messages = prompt_template_manager.render(name=f'ircot_{dataset}', prompt_user=prompt_user)

    try:
        response_message, metadata = llm_client.infer(messages=messages)
        response_content = response_message[0]["content"]
    except Exception as e:
        logger.exception("An exception occurred while calling LLM for QA!")
        return ''
    
    return response_content


from .base import BaseLLM
from ..utils.logging_utils import get_logger
from ..utils.config_utils import BaseConfig
from .request_llm import SiliconFlowLLM  # 替换原OpenAI/Bedrock导入

logger = get_logger(__name__)

def _get_llm_class(config: BaseConfig):
    # 仅返回硅基流动LLM实例
    return SiliconFlowLLM.from_experiment_config(config)
    


import re
import string

def normalize_answer(answer: str) -> str:
    """
    Normalize a given string by applying the following transformations:
    1. Convert the string to lowercase.
    2. Remove punctuation characters.
    3. Remove the articles "a", "an", and "the".
    4. Normalize whitespace by collapsing multiple spaces into one.

    Args:
        answer (str): The input string to be normalized.

    Returns:
        str: The normalized string.
    """
    def remove_articles(text):
        return re.sub(r"\b(a|an|the)\b", " ", text)

    def white_space_fix(text):
        return " ".join(text.split())

    def remove_punc(text):
        exclude = set(string.punctuation)
        return "".join(ch for ch in text if ch not in exclude)

    def lower(text):
        return text.lower()
    
    return white_space_fix(remove_articles(remove_punc(lower(answer))))



import os
from dataclasses import dataclass, field
from typing import (
    Literal,
    Union,
    Optional
)

from .logging_utils import get_logger

logger = get_logger(__name__)


@dataclass
class BaseConfig:
    """One and only configuration."""
    # LLM specific attributes 
    api_key: str = field(
        default="",
        metadata={"help": "You siliconflow api_key."}
    )
    llm_name: str = field(
        default="Pro/deepseek-ai/DeepSeek-V3",
        metadata={"help": "Class name indicating which LLM model to use."}
    )
    llm_base_url: str = field(
        default=None,
        metadata={"help": "Base URL for the LLM model, if none, means using OPENAI service."}
    )
    embedding_base_url: str = field(
        default=None,
        metadata={"help": "Base URL for an OpenAI compatible embedding model, if none, means using OPENAI service."}
    )
    azure_endpoint: str = field(
        default=None,
        metadata={"help": "Azure Endpoint URI for the LLM model, if none, uses OPENAI service directly."}
    )
    azure_embedding_endpoint: str = field(
        default=None,
        metadata={"help": "Azure Endpoint URI for the OpenAI embedding model, if none, uses OPENAI service directly."}
    )
    max_new_tokens: Union[None, int] = field(
        default=2048,
        metadata={"help": "Max new tokens to generate in each inference."}
    )
    num_gen_choices: int = field(
        default=1,
        metadata={"help": "How many chat completion choices to generate for each input message."}
    )
    seed: Union[None, int] = field(
        default=None,
        metadata={"help": "Random seed."}
    )
    temperature: float = field(
        default=0,
        metadata={"help": "Temperature for sampling in each inference."}
    )
    response_format: Union[dict, None] = field(
        default_factory=lambda: { "type": "json_object" },
        metadata={"help": "Specifying the format that the model must output."}
    )
    
    ## LLM specific attributes -> Async hyperparameters
    max_retry_attempts: int = field(
        default=5,
        metadata={"help": "Max number of retry attempts for an asynchronous API calling."}
    )
    # Storage specific attributes
    force_openie_from_scratch: bool = field(
        default=False,
        metadata={"help": "If set to True, will ignore all existing openie files and rebuild them from scratch."}
    )

    # Storage specific attributes 
    force_index_from_scratch: bool = field(
        default=False,
        metadata={"help": "If set to True, will ignore all existing storage files and graph data and will rebuild from scratch."}
    )
    rerank_dspy_file_path: str = field(
        default=None,
        metadata={"help": "Path to the rerank dspy file."}
    )
    passage_node_weight: float = field(
        default=0.05,
        metadata={"help": "Multiplicative factor that modified the passage node weights in PPR."}
    )
    save_openie: bool = field(
        default=True,
        metadata={"help": "If set to True, will save the OpenIE model to disk."}
    )
    
    # Preprocessing specific attributes
    text_preprocessor_class_name: str = field(
        default="TextPreprocessor",
        metadata={"help": "Name of the text-based preprocessor to use in preprocessing."}
    )
    preprocess_encoder_name: str = field(
        default="Pro/deepseek-ai/DeepSeek-V3",
        metadata={"help": "Name of the encoder to use in preprocessing (currently implemented specifically for doc chunking)."}
    )
    preprocess_chunk_overlap_token_size: int = field(
        default=128,
        metadata={"help": "Number of overlap tokens between neighbouring chunks."}
    )
    preprocess_chunk_max_token_size: int = field(
        default=None,
        metadata={"help": "Max number of tokens each chunk can contain. If set to None, the whole doc will treated as a single chunk."}
    )
    preprocess_chunk_func: Literal["by_token", "by_word"] = field(default='by_token')
    
    
    # Information extraction specific attributes
    information_extraction_model_name: Literal["openie_openai_gpt", ] = field(
        default="openie_openai_gpt",
        metadata={"help": "Class name indicating which information extraction model to use."}
    )
    openie_mode: Literal["offline", "online"] = field(
        default="online",
        metadata={"help": "Mode of the OpenIE model to use."}
    )
    
    # Embedding specific attributes
    embedding_model_name: str = field(
        default="nvidia/NV-Embed-v2",
        metadata={"help": "Class name indicating which embedding model to use."}
    )
    embedding_batch_size: int = field(
        default=16,
        metadata={"help": "Batch size of calling embedding model."}
    )
    embedding_return_as_normalized: bool = field(
        default=True,
        metadata={"help": "Whether to normalize encoded embeddings not."}
    )
    embedding_max_seq_len: int = field(
        default=2048,
        metadata={"help": "Max sequence length for the embedding model."}
    )
    embedding_model_dtype: Literal["float16", "float32", "bfloat16", "auto"] = field(
        default="auto",
        metadata={"help": "Data type for local embedding model."}
    )
    
    
    
    # Graph construction specific attributes
    synonymy_edge_topk: int = field(
        default=2047,
        metadata={"help": "k for knn retrieval in buiding synonymy edges."}
    )
    synonymy_edge_query_batch_size: int = field(
        default=1000,
        metadata={"help": "Batch size for query embeddings for knn retrieval in buiding synonymy edges."}
    )
    synonymy_edge_key_batch_size: int = field(
        default=10000,
        metadata={"help": "Batch size for key embeddings for knn retrieval in buiding synonymy edges."}
    )
    synonymy_edge_sim_threshold: float = field(
        default=0.8,
        metadata={"help": "Similarity threshold to include candidate synonymy nodes."}
    )
    is_directed_graph: bool = field(
        default=False,
        metadata={"help": "Whether the graph is directed or not."}
    )
    
    
    
    # Retrieval specific attributes
    linking_top_k: int = field(
        default=5,
        metadata={"help": "The number of linked nodes at each retrieval step"}
    )
    retrieval_top_k: int = field(
        default=200,
        metadata={"help": "Retrieving k documents at each step"}
    )
    damping: float = field(
        default=0.5,
        metadata={"help": "Damping factor for ppr algorithm."}
    )
    
    
    # QA specific attributes
    max_qa_steps: int = field(
        default=1,
        metadata={"help": "For answering a single question, the max steps that we use to interleave retrieval and reasoning."}
    )
    qa_top_k: int = field(
        default=5,
        metadata={"help": "Feeding top k documents to the QA model for reading."}
    )
    
    # Save dir (highest level directory)
    save_dir: str = field(
        default=None,
        metadata={"help": "Directory to save all related information. If it's given, will overwrite all default save_dir setups. If it's not given, then if we're not running specific datasets, default to `outputs`, otherwise, default to a dataset-customized output dir."}
    )
    
    
    
    # Dataset running specific attributes
    ## Dataset running specific attributes -> General
    dataset: Optional[Literal['hotpotqa', 'hotpotqa_train', 'musique', '2wikimultihopqa']] = field(
        default=None,
        metadata={"help": "Dataset to use. If specified, it means we will run specific datasets. If not specified, it means we're running freely."}
    )
    ## Dataset running specific attributes -> Graph
    graph_type: Literal[
        'dpr_only', 
        'entity', 
        'passage_entity', 'relation_aware_passage_entity',
        'passage_entity_relation', 
        'facts_and_sim_passage_node_unidirectional',
    ] = field(
        default="facts_and_sim_passage_node_unidirectional",
        metadata={"help": "Type of graph to use in the experiment."}
    )
    corpus_len: Optional[int] = field(
        default=None,
        metadata={"help": "Length of the corpus to use."}
    )
    
    
    def __post_init__(self):
        if self.save_dir is None: # If save_dir not given
            if self.dataset is None: self.save_dir = 'outputs' # running freely
            else: self.save_dir = os.path.join('outputs', self.dataset) # customize your dataset's output dir here
        logger.debug(f"Initializing the highest level of save_dir to be {self.save_dir}")



best_dspy_prompt = {
    'prog': {
        'lm': None,
        'traces': [],
        'train': [],
        'demos': [], # demos被直接去掉了，因为对函数调用用处不大
        # 关键修正：将原来的顶层'system'移到'prog'内部
        'system': '你的输入字段是：\n1. `question`（字符串）：用于检索的查询\n2. `fact_before_filter`（字符串）：待过滤的候选事实\n\n你的输出字段是：\n1. `fact_after_filter`（事实）：JSON格式的过滤后事实\n\n所有交互将按以下方式结构化，并填入适当的值。\n\n[[ ## question ## ]]\n{question}\n\n[[ ## fact_before_filter ## ]]\n{fact_before_filter}\n\n[[ ## fact_after_filter ## ]]\n{fact_after_filter}        # 注意：你生成的值必须符合以下JSON模式：{"type": "object", "properties": {"fact": {"type": "array", "description": "事实列表，每个事实是包含3个字符串的列表：[主体, 谓词, 客体]", "items": {"type": "array", "items": {"type": "string"}}, "title": "Fact"}}, "required": ["fact"], "title": "Fact"}\n\n[[ ## completed ## ]]\n\n遵循此结构时，你的目标是：\n        你是全球顶尖研究人员和决策者使用的高风险问答系统的关键组成部分。你的任务是根据与给定查询的相关性过滤事实，确保向这些利益相关者呈现最关键的信息。该查询需要仔细分析，并可能需要多步推理来连接不同的信息片段。你必须从提供的候选列表中选择最多4个相关事实，这些事实与查询有很强的关联，有助于推理并提供准确的答案。输出应采用JSON格式，例如{"fact": [["主体1", "谓词1", "客体1"], ["主体2", "谓词2", "客体2"]]}，如果没有相关事实，则返回空列表{"fact": []}。你的响应准确性至关重要，因为它将直接影响这些高层决策者的决策。你只能使用候选列表中的事实，不得生成新事实。关键决策的未来取决于你准确过滤和呈现相关信息的能力。'
    },
    'signature': {
        'instructions': '你是全球顶尖研究人员和决策者使用的高风险问答系统的关键组成部分。你的任务是根据与给定查询的相关性过滤事实，确保向这些利益相关者呈现最关键的信息。该查询需要仔细分析，并可能需要多步推理来连接不同的信息片段。你必须从提供的候选列表中选择最多4个相关事实，这些事实与查询有很强的关联，有助于推理并提供准确的答案。输出应采用JSON格式，例如{"fact": [["主体1", "谓词1", "客体1"], ["主体2", "谓词2", "客体2"]]}，如果没有相关事实，则返回空列表{"fact": []}。你的响应准确性至关重要，因为它将直接影响这些高层决策者的决策。你只能使用候选列表中的事实，不得生成新事实。关键决策的未来取决于你准确过滤和呈现相关信息的能力。',
        'fields': [
            {
                'prefix': '问题：',
                'description': '用于检索的查询'
            },
            {
                'prefix': '过滤前的事实：',
                'description': '待过滤的候选事实'
            },
            {
                'prefix': '过滤后的事实：',
                'description': 'JSON格式的过滤后事实'
            }
        ]
    }
    # 注意：原来的顶层'system'已移走，这里不再保留
}



from copy import deepcopy
from typing import List, Optional
import os

import numpy as np
import requests
from tqdm import tqdm

from ..utils.config_utils import BaseConfig
from ..utils.logging_utils import get_logger
from .base import BaseEmbeddingModel, EmbeddingConfig, make_cache_embed

logger = get_logger(__name__)

class Qwen3EmbeddingModel(BaseEmbeddingModel):
    def __init__(self, global_config: Optional[BaseConfig] = None, embedding_model_name: Optional[str] = None) -> None:
        super().__init__(global_config=global_config)
        # 支持通过参数覆盖模型名称（解决"unexpected keyword argument 'embedding_model_name'"错误）
        if embedding_model_name is not None:
            self.embedding_model_name = embedding_model_name
            logger.debug(f"Overriding embedding model name with: {self.embedding_model_name}")
        self._init_embedding_config()
        # 初始化硅基流动API客户端
        self._init_siliconflow_client()

    def _init_siliconflow_client(self) -> None:
        """初始化硅基流动API连接参数"""
        # 从全局配置或环境变量获取API密钥
        self.api_key = self.global_config.api_key
        if not self.api_key:
            raise ValueError("OPENAI_API_KEY must be set")
        # 硅基流动嵌入API端点
        self.api_url = self.global_config.embedding_base_url
        # 请求头部（固定格式）
        self.headers = {
            "Authorization": self.api_key,
            "Content-Type": "application/json"
        }
        logger.debug("Silicon Flow API client initialized")

    def _init_embedding_config(self) -> None:
        """初始化嵌入模型配置（参考OpenAI实现）"""
        config_dict = {
            "embedding_model_name": self.embedding_model_name,
            "norm": self.global_config.embedding_return_as_normalized,
            "model_init_params": {
                "pretrained_model_name_or_path": self.embedding_model_name,
                # 硅基流动API无需本地模型参数，此处保留结构用于兼容性
            },
            "encode_params": {
                "max_length": self.global_config.embedding_max_seq_len or 8192,  # Qwen3支持长文本
                "batch_size": self.global_config.embedding_batch_size or 16,
                "num_workers": 32
            },
        }
        self.embedding_config = EmbeddingConfig.from_dict(config_dict=config_dict)
        logger.debug(f"Qwen3 embedding config initialized: {self.embedding_config}")

    def encode(self, texts: List[str]) -> np.ndarray:
        """单批文本编码（适配硅基流动API格式）"""
        # 文本预处理（同OpenAI逻辑）
        texts = [t.replace("\n", " ") for t in texts]
        texts = [t if t != "" else " " for t in texts]
        
        # 构造硅基流动API请求体
        payload = {
            "model": self.embedding_model_name,  # 模型名称，如"Qwen/Qwen3-Embedding-8B"
            "input": texts  # 支持单文本或文本列表
        }
        
        # 发送POST请求
        try:
            response = requests.post(
                url=self.api_url,
                headers=self.headers,
                json=payload,
                timeout=30  # 超时设置
            )
            response.raise_for_status()  # 触发HTTP错误（如401、404）
        except requests.exceptions.RequestException as e:
            logger.error(f"Silicon Flow API request failed: {str(e)}")
            raise
        
        # 解析响应
        result = response.json()
        if "data" not in result:
            raise ValueError(f"Invalid response from Silicon Flow API: {result}")
        
        # 提取嵌入向量并转换为numpy数组
        embeddings = np.array([item["embedding"] for item in result["data"]])
        return embeddings

    def batch_encode(self, texts: List[str], **kwargs) -> np.ndarray:
        """批量文本编码（分片处理大列表）"""
        if isinstance(texts, str):
            texts = [texts]
        params = deepcopy(self.embedding_config.encode_params)
        if kwargs:
            params.update(kwargs)
        
        batch_size = params.pop("batch_size", 16)
        logger.debug(f"Batch encoding with batch size: {batch_size}")
        
        if len(texts) <= batch_size:
            results = self.encode(texts)
        else:
            # 大列表分片处理，显示进度条
            pbar = tqdm(total=len(texts), desc="Qwen3 Batch Encoding")
            results = []
            for i in range(0, len(texts), batch_size):
                batch = texts[i:i + batch_size]
                results.append(self.encode(batch))
                pbar.update(len(batch))
            pbar.close()
            results = np.concatenate(results)
        
        # 向量归一化（同OpenAI逻辑）
        if self.embedding_config.norm:
            results = (results.T / np.linalg.norm(results, axis=1)).T
        return results


import json
from dataclasses import dataclass, field, asdict
from typing import (
    Optional,
    Tuple,
    Any, 
    Dict,
    List
)
import numpy as np
import threading
import multiprocessing


from ..utils.logging_utils import get_logger
from ..utils.config_utils import BaseConfig


logger = get_logger(__name__)


@dataclass
class EmbeddingConfig:
    _data: Dict[str, Any] = field(default_factory=dict, init=False, repr=False)

    def __getattr__(self, key: str) -> Any:
        # Define patterns to ignore for Jupyter/IPython-related attributes
        ignored_prefixes = ("_ipython_", "_repr_")
        if any(key.startswith(prefix) for prefix in ignored_prefixes):
            raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{key}'")
    
        if key in self._data:
            return self._data[key]
        
        logger.error(f"'{self.__class__.__name__}' object has no attribute '{key}'")
        raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{key}'")


    def __setattr__(self, key: str, value: Any) -> None:
        if key == '_data':
            super().__setattr__(key, value)
        else:
            self._data[key] = value

    def __delattr__(self, key: str) -> None:
        if key in self._data:
            del self._data[key]
        else:
            logger.error(f"'{self.__class__.__name__}' object has no attribute '{key}'")
            raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{key}'")

    def __getitem__(self, key: str) -> Any:
        """Allow dict-style key lookup."""
        if key in self._data:
            return self._data[key]
        logger.error(f"'{key}' not found in configuration.")
        raise KeyError(f"'{key}' not found in configuration.")

    def __setitem__(self, key: str, value: Any) -> None:
        """Allow dict-style key assignment."""
        self._data[key] = value

    def __delitem__(self, key: str) -> None:
        """Allow dict-style key deletion."""
        if key in self._data:
            del self._data[key]
        else:
            logger.error(f"'{key}' not found in configuration.")
            raise KeyError(f"'{key}' not found in configuration.")

    def __contains__(self, key: str) -> bool:
        """Allow usage of 'in' to check for keys."""
        return key in self._data
    
    
    def batch_upsert(self, updates: Dict[str, Any]) -> None:
        """Update existing attributes or add new ones from the given dictionary."""
        self._data.update(updates)

    def to_dict(self) -> Dict[str, Any]:
        """Export the configuration as a JSON-serializable dictionary."""
        return self._data

    def to_json(self) -> str:
        """Export the configuration as a JSON string."""
        return json.dumps(self._data)
    
    @classmethod
    def from_dict(cls, config_dict: Dict[str, Any]) -> "LLMConfig":
        """Create an LLMConfig instance from a dictionary."""
        instance = cls()
        instance.batch_upsert(config_dict)
        return instance

    @classmethod
    def from_json(cls, json_str: str) -> "LLMConfig":
        """Create an LLMConfig instance from a JSON string."""
        instance = cls()
        instance.batch_upsert(json.loads(json_str))
        return instance

    def __str__(self) -> str:
        """Provide a user-friendly string representation of the configuration."""
        return json.dumps(self._data, indent=4)
    

from filelock import FileLock
import sqlite3
import hashlib
import os
def make_cache_embed(encode_func, cache_file_name, device):
    def wrapper(**kwargs):
        # FOCUS_KEYS = ["instruction", "prompts", "max_length"]
        instruction = kwargs.get("instruction", "")
        max_length = kwargs.get("max_length", "")

        hash_strs = []
        for prompt in kwargs['prompts']:
            key_str = json.dumps({
                "instruction": instruction,
                "promps": prompt,
                "max_length": max_length
            }, sort_keys=True, default=str)
            hash_strs.append(hashlib.sha256(key_str.encode("utf-8")).hexdigest())

        missed_prompts = []
        lock_file = cache_file_name + ".lock"

        with FileLock(lock_file):
            with sqlite3.connect(cache_file_name) as conn:
                cursor = conn.cursor()
                cursor.execute('''
                    CREATE TABLE IF NOT EXISTS embeddings (
                        hash TEXT PRIMARY KEY,
                        embedding BLOB
                    )
                ''')
                conn.commit()

                embeddings = []
                for i, hash_str in enumerate(hash_strs):
                    cursor.execute('SELECT embedding FROM embeddings WHERE hash = ?', (hash_str,))
                    result = cursor.fetchone()
                    if result:
                        # Convert the BLOB back to a NumPy array
                        emb = np.frombuffer(result[0], dtype=np.float32)
                        embeddings.append(emb)
                    else:
                        missed_prompts.append(i)
                        embeddings.append(None)

        if missed_prompts:
            # Update kwargs to include only the missed prompts.
            kwargs['prompts'] = [kwargs['prompts'][i] for i in missed_prompts]
            # 下列torch数据结构均使用numpy类似数据结构替换，但原注释暂且保留
            # Call the encoder function (which returns a 2D torch tensor)
            new_embeddings = encode_func(**kwargs)
            # Insert the new embeddings back into the correct positions.
            for idx, embedding in enumerate(new_embeddings):
                embeddings[missed_prompts[idx]] = embedding

            # Save the new embeddings to the cache.
            with FileLock(lock_file):
                with sqlite3.connect(cache_file_name) as conn:
                    cursor = conn.cursor()
                    for i in missed_prompts:
                        hash_str = hash_strs[i]
                        emb = embeddings[i]
                        # Convert torch tensor to numpy bytes if necessary.
                        if isinstance(emb, np.ndarray):
                            emb_bytes = emb.tobytes()
                        else:
                            raise TypeError({type(emb)})
                        cursor.execute('INSERT INTO embeddings (hash, embedding) VALUES (?, ?)', (hash_str, emb_bytes))
                    conn.commit()

        # Convert all embeddings to torch tensors if they're not already
        final_embeddings = [emb for emb in final_embeddings]

        # Return a 2D tensor where each row is an embedding.
        return np.stack(final_embeddings)

    return wrapper
    
class BaseEmbeddingModel:
    global_config: BaseConfig
    embedding_model_name: str # Class name indicating which embedding model to use.
    embedding_config: EmbeddingConfig
    
    embedding_dim: int # Need subclass to init
    
    def __init__(self, global_config: Optional[BaseConfig] = None) -> None:
        if global_config is None: 
            logger.debug("global config is not given. Using the default ExperimentConfig instance.")
            self.global_config = BaseConfig()
        else: self.global_config = global_config
        logger.debug(f"Loading {self.__class__.__name__} with global_config: {asdict(self.global_config)}")
        
        
        self.embedding_model_name = self.global_config.embedding_model_name

        logger.debug(f"Init {self.__class__.__name__}'s embedding_model_name with: {self.embedding_model_name}")

    def batch_encode(self, texts: List[str], **kwargs) -> None:
        raise NotImplementedError
    
    
    def get_query_doc_scores(self, query_vec: np.ndarray, doc_vecs: np.ndarray):
        # """
        # @param query_vec: query vector
        # @param doc_vecs: doc matrix
        # @return: a matrix of query-doc scores
        # """
        return np.dot(query_vec, doc_vecs.T)
    


class EmbeddingCache:
    """A multiprocessing-safe global cache for storing embeddings."""
    
    _manager = multiprocessing.Manager()
    _cache = _manager.dict()  # Shared dictionary for multiprocessing
    _lock = threading.Lock()  # Thread-safe lock for concurrent access

    @classmethod
    def get(cls, content):
        """Retrieve the embedding if cached."""
        return cls._cache.get(content)

    @classmethod
    def set(cls, content, embedding):
        """Store an embedding in the cache."""
        with cls._lock:  # Ensures thread safety
            cls._cache[content] = embedding

    @classmethod
    def contains(cls, content):
        """Check if the embedding exists in cache."""
        return content in cls._cache

    @classmethod
    def clear(cls):
        """Clear the entire cache."""
        with cls._lock:
            cls._cache.clear()


ner_system = """你的任务是从给定段落中提取命名实体。
请以JSON列表的形式返回实体。
"""

one_shot_ner_paragraph = """Radio City
Radio City是印度第一家私人调频电台，于2001年7月3日开播。
它播放印地语、英语和地方语言歌曲。
Radio City最近于2008年5月进军新媒体领域，推出了音乐门户网站——PlanetRadiocity.com，该网站提供音乐相关新闻、视频、歌曲和其他音乐相关功能。"""


one_shot_ner_output = """{"named_entities":
    ["Radio City", "印度", "2001年7月3日", "印地语", "英语", "2008年5月", "PlanetRadiocity.com"]
}
"""


prompt_template = [
    {"role": "system", "content": ner_system},
    {"role": "user", "content": one_shot_ner_paragraph},
    {"role": "assistant", "content": one_shot_ner_output},
    {"role": "user", "content": "${passage}"}
]


from .base import EmbeddingConfig, BaseEmbeddingModel
from .Qwen3 import Qwen3EmbeddingModel

from ..utils.logging_utils import get_logger

logger = get_logger(__name__)


def _get_embedding_model_class(embedding_model_name: str = "nvidia/NV-Embed-v2"):
    if "Qwen/Qwen3-Embedding-8B" in embedding_model_name:
        return Qwen3EmbeddingModel
    assert False, f"Unknown embedding model name: {embedding_model_name}"


import json
from abc import ABC, abstractmethod
from dataclasses import dataclass, field, asdict
from typing import (
    Optional,
    Tuple,
    Any, 
    Dict,
    List
)


from ..utils.logging_utils import get_logger
from ..utils.config_utils import BaseConfig
from ..utils.llm_utils import (
    TextChatMessage
)



logger = get_logger(__name__)




@dataclass
class LLMConfig:
    _data: Dict[str, Any] = field(default_factory=dict, init=False, repr=False)

    def __getattr__(self, key: str) -> Any:
        # Define patterns to ignore for Jupyter/IPython-related attributes
        ignored_prefixes = ("_ipython_", "_repr_")
        if any(key.startswith(prefix) for prefix in ignored_prefixes):
            raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{key}'")
    
        if key in self._data:
            return self._data[key]
        
        logger.error(f"'{self.__class__.__name__}' object has no attribute '{key}'")
        raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{key}'")


    def __setattr__(self, key: str, value: Any) -> None:
        if key == '_data':
            super().__setattr__(key, value)
        else:
            self._data[key] = value

    def __delattr__(self, key: str) -> None:
        if key in self._data:
            del self._data[key]
        else:
            logger.error(f"'{self.__class__.__name__}' object has no attribute '{key}'")
            raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{key}'")

    def __getitem__(self, key: str) -> Any:
        """Allow dict-style key lookup."""
        if key in self._data:
            return self._data[key]
        logger.error(f"'{key}' not found in configuration.")
        raise KeyError(f"'{key}' not found in configuration.")

    def __setitem__(self, key: str, value: Any) -> None:
        """Allow dict-style key assignment."""
        self._data[key] = value

    def __delitem__(self, key: str) -> None:
        """Allow dict-style key deletion."""
        if key in self._data:
            del self._data[key]
        else:
            logger.error(f"'{key}' not found in configuration.")
            raise KeyError(f"'{key}' not found in configuration.")

    def __contains__(self, key: str) -> bool:
        """Allow usage of 'in' to check for keys."""
        return key in self._data
    
    
    def batch_upsert(self, updates: Dict[str, Any]) -> None:
        """Update existing attributes or add new ones from the given dictionary."""
        self._data.update(updates)

    def to_dict(self) -> Dict[str, Any]:
        """Export the configuration as a JSON-serializable dictionary."""
        return self._data

    def to_json(self) -> str:
        """Export the configuration as a JSON string."""
        return json.dumps(self._data)
    
    @classmethod
    def from_dict(cls, config_dict: Dict[str, Any]) -> "LLMConfig":
        """Create an LLMConfig instance from a dictionary."""
        instance = cls()
        instance.batch_upsert(config_dict)
        return instance

    @classmethod
    def from_json(cls, json_str: str) -> "LLMConfig":
        """Create an LLMConfig instance from a JSON string."""
        instance = cls()
        instance.batch_upsert(json.loads(json_str))
        return instance

    def __str__(self) -> str:
        """Provide a user-friendly string representation of the configuration."""
        return json.dumps(self._data, indent=4)




class BaseLLM(ABC):
    """Abstract base class for LLMs."""
    global_config: BaseConfig
    llm_name: str # Class name indicating which LLM model to use.
    llm_config: LLMConfig  # Store LLM specific config, init and handled by specifc LLM
    
    
    def __init__(self, global_config: Optional[BaseConfig] = None) -> None:
        if global_config is None: 
            logger.debug("global config is not given. Using the default ExperimentConfig instance.")
            self.global_config = BaseConfig()
        else: self.global_config = global_config
        logger.debug(f"Loading {self.__class__.__name__} with global_config: {asdict(self.global_config)}")
        
        self.llm_name = self.global_config.llm_name
        logger.debug(f"Init {self.__class__.__name__}'s llm_name with: {self.llm_name}")


    @abstractmethod
    def _init_llm_config(self) -> None:
        """
        Each LLM model should extract its own running parameters from global_config and raise exception if any mandatory parameter is not defined in global_config.
        This function must init `self.llm_config`.
        """
        pass
    
    
    def batch_upsert_llm_config(self, updates: Dict[str, Any]) -> None:
        """
        Upsert self.llm_config with attribute-value pairs specified by a dict. 
        
        Args:
            updates (Dict[str, Any]): a dict to be integrated into self.llm_config.
            
        Returns: 
            None
        """
        
        self.llm_config.batch_upsert(updates=updates)
        logger.debug(f"Updated {self.__class__.__name__}'s llm_config with {updates} to eventually obtain llm_config as: {self.llm_config}")
    
    
    def ainfer(self, chat: List[TextChatMessage]) -> Tuple[List[TextChatMessage], dict]:
        """
        Perform asynchronous inference using the LLM.
        
        Args:
            chat (List[TextChatMessage]): Input chat history for the LLM.

        Returns:
            Tuple[List[TextChatMessage], dict]: The list of n (number of choices) LLM response message (a single dict of role + content), and additional metadata (all input params including input chat) as a dictionary.
        """
        pass
    

 
    def infer(self, chat: List[TextChatMessage]) -> Tuple[List[TextChatMessage], dict]:
        """
        Perform synchronous inference using the LLM.
        
        Args:
            chat (List[TextChatMessage]): Input chat history for the LLM.

        Returns:
            Tuple[List[TextChatMessage], dict]: The list of n (number of choices) LLM response message (a single dict of role + content), and additional metadata (all input params including input chat) as a dictionary.
        """
        pass
    


    def batch_infer(self, batch_chat: List[List[TextChatMessage]]) -> Tuple[List[List[TextChatMessage]], List[dict]]:
        """
        Perform batched synchronous inference using the LLM.
        
        Args:
            batch_chat (List[List[TextChatMessage]]): Input chat history batch for the LLM.

        Returns:
            Tuple[List[List[TextChatMessage]], List[dict]]: The batch list of length-n (number of choices) list of LLM response message (a single dict of role + content), and corresponding batch of additional metadata (all input params including input chat) as a list of dictionaries.
        """
        
        pass
        
        
# # Example usage
# if __name__ == "__main__":
#     config = LLMConfig()
#     config.batch_upsert({"learning_rate": 0.001, "batch_size": 32})
#     print(config.to_dict())

#     config.optimizer = "adam"
#     print(config.to_dict())

#     json_config = config.to_json()
#     print(json_config)

#     new_config = LLMConfig.from_json(json_config)
#     print(new_config.to_dict())

#     dict_config = {"dropout": 0.5, "epochs": 10}
#     another_config = LLMConfig.from_dict(dict_config)
#     print(another_config.to_dict())
